rem
	Character Maker
	~ BCC Edition ~
				
				ZaP
end rem

SuperStrict
Import maxgui.drivers

' const
Const apptitle:String = "ZaP's Character Maker [BCC Edition]"

' type
Type TCharacter
	Field bodypart:TMap
	Field bodypartr:TMap
	Field bodypartg:TMap
	Field bodypartb:TMap
	Field imageset:String
	
	Method New()
		Self.bodypart = New TMap
		Self.bodypartr = New TMap
		Self.bodypartg = New TMap
		Self.bodypartb = New TMap
		
		Local b:String
		
		For b = EachIn MapKeys(imagemap)
			Self.bodypartr.Insert(b, "255")
			Self.bodypartg.Insert(b, "255")
			Self.bodypartb.Insert(b, "255")
		Next
	End Method
	
	Method SetImageAssocs(m:TMap)
		Local key:String
		
		For key = EachIn MapKeys(m)
			Self.bodypart.Insert(key, "0")
		Next
	End Method
	
	Method Draw()
		Local i:Int = 0
		Local j:TJoint

		SetOrigin(GraphicsWidth() / 2 - (tilesize / 2 * scale), GraphicsHeight() / 2 - (tilesize / 2 * scale))
		SetScale(scale, scale)
		
		For j:TJoint = EachIn currentframe.joints
			If j.parent <> Null And j.name <> ""
				SetColor(Int(String(bodypartr.ValueForKey(j.name))), Int(String(bodypartg.ValueForKey(j.name))), Int(String(bodypartb.ValueForKey(j.name))))
				DrawImageAligned(TImage(imagemap.ValueForKey(j.name)), j, Int(String(bodypart.ValueForKey(j.name))))
			EndIf
		Next
	End Method
	
	Method Export(file:String)
		If file = "" Then Return
		
		Local i:Int = 0
		Local maxw:Int = 0
		
		' determine image width (longest animation)
		For i = 0 To CountGadgetItems(animlist) - 1
			Local animlen:Int = TAnim(GadgetItemExtra(animlist, i)).framelist.Count()
			If animlen > maxw Then maxw = animlen
		Next
	
		SetClsColor(255, 0, 255)
		Cls
		
		Local out:TPixmap = CreatePixmap(tilesize * maxw, tilesize * CountGadgetItems(animlist), PF_RGBA8888)
		Local x:Int = 0
		Local y:Int = 0
		
		' draw each animation on a horizontal stripe of the target pixmap
		For i = 0 To CountGadgetItems(animlist) - 1
			Local anim:TAnim = TAnim(GadgetItemExtra(animlist, i))
			Local f:TFrame = Null
			Local temppixmap:TPixmap = Null
			
			currentanimation = anim
			x = 0
			
			For f = EachIn anim.framelist
				Cls
				SetFrame(f.num)
				Self.Draw()
				temppixmap = GrabPixmap(GraphicsWidth() / 2 - (tilesize / 2), GraphicsHeight() / 2 - (tilesize / 2), tilesize, tilesize)
				out.Paste(temppixmap, x * tilesize, y * tilesize)
				x:+1
			Next
			
			y:+1
		Next
		
		SavePixmapPNG(out, file)
		Notify("Export successful.")
		SAVED = True
	End Method
End Type

Type TJoint
	Field x:Float
	Field y:Float
	Field id:Int
	Field parent:TJoint
	Field frame:TFrame
	Field name:String
End Type

Type TFrame
	Field num:Int
	Field joints:TList
	
	Method New:TFrame()
		Self.joints = New TList
	End Method
End Type

Type TAnim
	Field framelist:TList
	Field name:String
	
	Method New:TAnim()
		Self.framelist = New TList
	End Method
End Type

' vip's
Local currentbodypart:String = ""
Local animspeed:Int = 200 ' ms to wait between frames
Local lastframeupdate:Int = 0
Local animtimer:TTimer = CreateTimer(60)

Global currentchar:TCharacter
Global SAVED:Byte = True
Global currentframe:TFrame = Null
Global currentanimation:TAnim = Null
Global scale:Byte = 1
Global tilesize:Int = 0
Global imagemap:TMap = New TMap ' body part descriptions vs image handles
Global imagecount:TMap = New TMap ' animation frames

' Resource management
Try
	' load image associations and animations
	LoadImageSet("default")
	currentanimation = LoadAnimation("default")
	currentframe = TFrame(currentanimation.framelist.ValueAtIndex(0))
Catch loaderror:String
	Notify(loaderror, True)
	End
EndTry

currentchar = New TCharacter

' GUI
Local windowconfig:Int = WINDOW_TITLEBAR | WINDOW_CENTER | WINDOW_STATUS | WINDOW_RESIZABLE | WINDOW_CLIENTCOORDS
Local win:TGadget = CreateWindow(apptitle, 0, 0, 640, 480, Null, windowconfig)

Local canvas:TGadget = CreateCanvas(210, 5, ClientWidth(win) - 210, ClientHeight(win) - 10, win)
SetGraphics(CanvasGraphics(canvas))
SetBlend(ALPHABLEND)

Local viewpanel:TGadget = CreatePanel(5, 5, 200, 55, win, PANEL_GROUP, "Display options")
SetGadgetLayout(viewpanel, EDGE_ALIGNED, EDGE_CENTERED, EDGE_ALIGNED, EDGE_CENTERED)

Local zoomin:TGadget = CreateButton("Zoom +", ClientWidth(viewpanel) / 2, 5, ClientWidth(viewpanel) / 2 - 5, 24, viewpanel)
Local zoomout:TGadget = CreateButton("Zoom -", 5, 5, ClientWidth(viewpanel) / 2 - 5, 24, viewpanel)

Local toolspanel:TGadget = CreatePanel(5, 70, 200, 130, win, PANEL_GROUP, "Character Options")
SetGadgetLayout(toolspanel, EDGE_ALIGNED, EDGE_CENTERED, EDGE_ALIGNED, EDGE_CENTERED)

Global bodypartselector:TGadget = CreateComboBox(5, 5, ClientWidth(toolspanel) - 10 - 40, 24, toolspanel)

Local name:String = ""
For name = EachIn MapKeys(imagemap)
	AddGadgetItem(bodypartselector, name)
Next

SelectGadgetItem(bodypartselector, 0)

Local bpvaluestepper:TGadget = CreateSlider(ClientWidth(toolspanel) - 10 - 34, 5, 44, 22, toolspanel, SLIDER_STEPPER | SLIDER_HORIZONTAL)
Local colorbutton:TGadget = CreateButton("Color", 5, 40, ClientWidth(toolspanel) - 10 - 30, 24, toolspanel)
Local colorpanel:TGadget = CreatePanel(ClientWidth(toolspanel) - 10 - 20, 40, 24, 24, toolspanel, PANEL_SUNKEN)
Local imageset:TGadget = CreateButton("Change image set...", 5, 85, ClientWidth(toolspanel) - 10, 24, toolspanel)

Local animpanel:TGadget = CreatePanel(5, GadgetY(toolspanel) + GadgetHeight(toolspanel) + 10, 200, 230, win, PANEL_GROUP, "Animation")
SetGadgetLayout(animpanel, EDGE_ALIGNED, EDGE_CENTERED, EDGE_ALIGNED, EDGE_CENTERED)

Global animlist:TGadget = CreateListBox(5, 5, ClientWidth(animpanel) - 10, 60, animpanel)
Local addanim:TGadget = CreateButton("Add anim....", 5, 66, ClientWidth(animpanel) / 2 - 5, 24, animpanel)
Local delanim:TGadget = CreateButton("Delete anim.", ClientWidth(animpanel) / 2, 66, ClientWidth(animpanel) / 2 - 5, 24, animpanel)
Local prevframe:TGadget = CreateButton("< Prev. frame", 5, 100, ClientWidth(animpanel) / 2 - 5, 24, animpanel)
Local nextframe:TGadget = CreateButton("Next frame >", ClientWidth(animpanel) / 2, 100, ClientWidth(animpanel) / 2 - 5, 24, animpanel)
Local playanim:TGadget = CreateButton("Play", 5, 130, ClientWidth(animpanel) / 2 - 5, 24, animpanel)
Local stopanim:TGadget = CreateButton("Stop", ClientWidth(animpanel) / 2, 130, ClientWidth(animpanel) / 2 - 5, 24, animpanel)
CreateLabel("Animation speed:", 5, 165, ClientWidth(animpanel) - 10, 16, animpanel)
Local animspeedslider:TGadget = CreateSlider(5, 180, ClientWidth(animpanel) - 10, 24, animpanel, SLIDER_TRACKBAR | SLIDER_HORIZONTAL)
SetSliderRange(animspeedslider, 0, 999)
AddGadgetItem(animlist, "default", 0, -1, "", Object(currentanimation))

Local export:TGadget = CreateButton("Export...", 5, GadgetY(animpanel) + GadgetHeight(animpanel) + 10, 200, 24, win)
SetGadgetLayout(export, EDGE_ALIGNED, EDGE_CENTERED, EDGE_ALIGNED, EDGE_CENTERED)

' event handling
Repeat
	Select WaitEvent()
		Case EVENT_WINDOWCLOSE, EVENT_APPTERMINATE
			If SAVED
				Exit
			ElseIf Confirm("Character has not been saved. Exit?", True)
				Exit
			EndIf
							
		Case EVENT_GADGETSELECT
			If EventSource() = animlist And SelectedGadgetItem(animlist) >= 0
				currentanimation = TAnim(EventExtra())
			EndIf
			
		Case EVENT_GADGETACTION
			Select EventSource()
				Case bodypartselector
					currentbodypart = GadgetItemText(bodypartselector, SelectedGadgetItem(bodypartselector))
					SetSliderRange(bpvaluestepper, 0, Int(String(imagecount.ValueForKey(currentbodypart))) - 1)
					SetSliderValue(bpvaluestepper, Int(String(currentchar.bodypart.ValueForKey(currentbodypart))))

					colorpanel.SetColor(Int(String(currentchar.bodypartr.ValueForKey(currentbodypart))), Int(String(currentchar.bodypartg.ValueForKey(currentbodypart))), Int(String(currentchar.bodypartb.ValueForKey(currentbodypart))))
					
				Case bpvaluestepper
					SAVED = False
					currentchar.bodypart.Remove(currentbodypart)
					currentchar.bodypart.Insert(currentbodypart, String(SliderValue(bpvaluestepper)))
															
				Case colorbutton, colorpanel
					Local color:Int = RequestColor(255, 255, 255)
					Local r:Byte = RequestedRed()
					Local g:Byte = RequestedGreen()
					Local b:Byte = RequestedBlue()
					
					currentchar.bodypartr.Remove(currentbodypart)
					currentchar.bodypartr.Insert(currentbodypart, String(r))
					
					currentchar.bodypartg.Remove(currentbodypart)
					currentchar.bodypartg.Insert(currentbodypart, String(g))
					
					currentchar.bodypartb.Remove(currentbodypart)
					currentchar.bodypartb.Insert(currentbodypart, String(b))
					
					colorpanel.SetColor(r, g, b)
					
				Case colorbutton, colorpanel
										
				Case zoomin
					scale:+1
					RedrawGadget(canvas)
					
				Case zoomout
					If scale > 1 Then scale:-1
					RedrawGadget(canvas)
					
				Case prevframe
					SetFrame(currentframe.num - 1)
					RedrawGadget(canvas)
				
				Case nextframe
					SetFrame(currentframe.num + 1)
					RedrawGadget(canvas)
					
				Case playanim
					animspeed = 500
					
				Case stopanim
					animspeed = -1
					
				Case animspeedslider
					animspeed = 1000 - SliderValue(animspeedslider)
					
				Case addanim
					Local animfile:String = StripDir(RequestFile("Load animation...", "", False, "animations/"))
					Local loadedanimation:TAnim
					If animfile <> "" Then loadedanimation:TAnim = LoadAnimation(animfile)
					If loadedanimation <> Null Then AddGadgetItem(animlist, loadedanimation.name, 0, -1, "", Object(loadedanimation))
					
				Case export
					currentchar.export(RequestFile("Export to PNG...", "PNG Files:png", True))
				
				Case imageset
					ChangeImageSet(StripDir(RequestDir("Select desired image set directory...", "imagesets/")))
					
				Case delanim
					If SelectedGadgetItem(animlist) >= 0 And CountGadgetItems(animlist) > 1
						animlist.RemoveItem(animlist.SelectedItem())
					Else
						Notify("At least one animation is required!")
					End If
					
			End Select
			
			RedrawGadget(canvas)
		
		Case EVENT_TIMERTICK
			If animspeed > 0
				If MilliSecs() - lastframeupdate >= animspeed
					lastframeupdate = MilliSecs()
					SetFrame(currentframe.num + 1)
					RedrawGadget(canvas)
				EndIf
			End If
			
		Case EVENT_GADGETPAINT
			If EventSource() = canvas
				Cls
				DrawBG()
				currentchar.Draw()
				Flip
			End If
	End Select
Forever

End


Function LoadImageSet(setname:String)
	Local dir:String = "imagesets/" + setname
	
	If FileType(dir) = FILETYPE_DIR
		' read tilesize from setinfo
		Local s:TStream = ReadFile(dir + "/setinfo")
		If Not s Then Throw("Invalid image set '" + setname + "': Setinfo missing!")
		tilesize = ReadByte(s)
		
		' read body part description vs image file name associations
		' and load corresponding images
		While Not s.Eof()
			Local descriptor:String = s.ReadLine()
			Local filename:String = s.ReadLine()
			Local framecount:String = Int(s.ReadLine())
			Local image:TImage = LoadAnimImage2("imagesets/" + setname + "/" + filename, Int(framecount), 0)
			If Not image Then Throw("Could not load image '" + filename + "'!")
			
			imagemap.Insert(descriptor, image)
			imagecount.Insert(descriptor, framecount)
		Wend
		
		s.Close()
	Else
		Throw("Image set not found '" + setname + "'!")
	EndIf
End Function

Function ChangeImageSet(setname:String)
	ClearGadgetItems(bodypartselector)
	imagemap.Clear()
	imagecount.Clear()
	
	LoadImageSet(setname)
	
	currentchar = New TCharacter
	
	Local name:String = ""
	For name = EachIn MapKeys(imagemap)
		AddGadgetItem(bodypartselector, name)
	Next
	
	SelectGadgetItem(bodypartselector, 0)
End Function

Function DrawImageAligned(img:TImage, joint:TJoint, frame:Int = 0)
	If img = Null Then Return
	Local jointlen:Float = Sqr(((joint.x * tilesize - joint.parent.x * tilesize) ^ 2) + ((joint.y * tilesize - joint.parent.y * tilesize) ^ 2)) * scale
	Local angle:Float = ATan2(joint.x - joint.parent.x, joint.y - joint.parent.y)
		
	SetImageHandle(img, 0.0, img.height / 2.0 * scale)
	SetScale(jointlen / Float(img.width), jointlen / Float(img.width))
	SetRotation(90 - angle)
	
	Local tilew:Int = img.width / Float(String(imagecount.ValueForKey(joint.name)))
	DrawImage(img, joint.parent.x * tilesize, joint.parent.y * tilesize, frame)
End Function

Function LoadAnimation:TAnim(file:String)
	Local a:TAnim = New TAnim
	a.name = file
	
	Local s:TStream = ReadFile("animations/" + file)
	
	If Not s Then Throw("Could not find animation file '" + file + "'!")
	
	' read frame count + joint count
	Local fourtytwo:Byte = s.ReadByte()
	Local framecount:Int = s.ReadShort()
	Local jointcount:Int = s.ReadShort()
	
	Local i:Int = 0
	
	For i = 0 To framecount - 1
		Local f:TFrame = New TFrame
		f.num = i
		Local e:Int = 0
		
		For e = 0 To jointcount - 1
			Local j:TJoint = New TJoint
			j.frame = f
			j.id = s.ReadShort()
			Local parentid:Int = s.ReadShort()
			j.name = s.ReadString(s.ReadInt())
			j.x = s.ReadFloat()
			j.y = s.ReadFloat()
			j = AttachToParent(j, f, parentid)
			f.joints.AddLast(j)
		Next
			
		a.framelist.AddLast(f)
	Next
	
	s.Close()
	
	Return a
End Function

Function AttachToParent:TJoint(j:TJoint, f:TFrame, parentid:Int)
	If parentid >= 0
		Local p:TJoint
		For p:TJoint = EachIn f.joints
			If p.id = parentid
				j.parent = p
				Return j
			End If
		Next
	EndIf
	
	Return j
End Function

Function DrawBG()
	SetScale(1.0, 1.0)
	SetRotation(0.0)
	SetOrigin(0, 0)
	
	Local x:Int, y:Int
	
	For y = 0 To GraphicsHeight() Step 5
		For x = 0 To GraphicsWidth() Step 5
			If x Mod 2 = y Mod 2
				SetColor(60, 60, 60)
			Else
				SetColor(50, 50, 50)
			End If
			
			DrawRect(x, y, 5, 5)
		Next
	Next
	
	SetColor(255, 255, 255)
End Function

Function SetFrame(num:Int)
	If num > currentanimation.framelist.Count() - 1 Then num = num Mod currentanimation.framelist.Count()
	If num < 0 Then num = currentanimation.framelist.Count() - 1	
	currentframe = TFrame(currentanimation.framelist.ValueAtIndex(num))
End Function

Function LoadAnimImage2:TImage(url:String, framecount:Int, flags:Int = 0)
	Local img:TImage = LoadImage(url, flags)
	Local tilew:Int = img.width / framecount
	Local tileh:Int = img.height
	Return LoadAnimImage(img.pixmaps[0], tilew, tileh, 0, framecount, flags)
End Function